### timezones.R ###
#
# File to create a netCDF map of the world time zones. The offset relative to UTC is 
# mapped onto the grid specified below. The following input data are used:
# tz_world_mp: shape file of the time zones from http://efele.net/maps/tz/world/. The
# shape file contains the 400+ time zones from the tz database. This data are used to 
# create a gridded map of the time zones.
# tz_zones.txt: ascii file listing all time zones and the corresponding UTC offset. This
# table is used to assign the correct UTC offset to every time zone. The list was created
# from http://en.wikipedia.org/wiki/List_of_tz_database_time_zones.
# 
# HISTORY:
# 26 Jan 2015 - C. Keller - Initial version
#
#######################################################################################

# load requirements
library(rgdal)
library(raster)
library(ncdf4)

wdir <- "/home/ckeller/data/TIMEZONES/R"
setwd(wdir)

#######################################################################################
# timezones
#######################################################################################
if(FALSE){

	# define variables
	file      <- "timezones_1x1.nc"
	varname   <- "UTC_OFFSET"
	overwrite <- FALSE
	nlon      <- 360
	nlat      <- 180

	shp_dir   <- "data/tz_world_mp"
	layer     <- "tz_world_mp"
	
	tz_file   <- "data/tz_zones.txt"

	# create raster
	rast <- raster(nrows=nlat, ncols=nlon, xmn=-180, xmx=180, ymn=-90, ymx=90, 
				crs="+proj=longlat +datum=WGS84")

	# read time zone mapping table
	tz <- read_tz( tz_file )

	# Read shape file and pass to raster
	rastered <- readraster_shp( shp_dir, layer, rast, tz )
	
	# Write to netCDF
	write2ncdf( rast, rastered, file, varname=varname, add=!overwrite )
}

#######################################################################################
# readraster_shp
# # read shape file and put on raster. Use country ID's from the provided lookup table.
#######################################################################################
readraster_shp <- function(shp_dir, layer, rast, tz){
	
	# read shape file
	cat("reading shape file",shp_dir,"using layer",layer,"\n")
	shp <- try(readOGR(shp_dir, layer))
	if(class(shp)=="try-error"){
		cat(paste("Error in reading layer ", layer,"\n"))
		return(NULL)
	}

	# map country IDs
	IDs <- tz$UTC_OFFSET[match(shp$TZID, tz$ZONE_NAME)]
	if(any(is.na(IDs))){
		notFoundNames <- shp$TZID[which(is.na(IDs))]
		cat("cannot find utc offset for zone", as.character(unique(notFoundNames)), "\n")
		IDs[which(is.na(IDs))] <- -999.0
	}
	
	# put on raster
	cat("now rasterize...\n")
	rastered <- rasterize(shp, rast, IDs, background=-999.0)
	
	# all done
	return(rastered)
}

#######################################################################################
# read_tz
# read the time zones mapping table
#######################################################################################
read_tz <- function(tz_file){

	# open file and read every line
	fid <- file(tz_file)
	buff <- readLines(fid)
	close(fid)
	
	# extract zone names and utc shift in hours
	zone_name=character()
	utc_offset=real()
	for(ii in 1:length(buff)){
		
		# get line
		tmpline <- buff[ii]
		
		# split line by tabs
		splits <- strsplit(tmpline, split="\t")[[1]]
		
		# time zone is the third entry
		zone_name <- c(zone_name, trim(splits[3]))
		
		# utc offset is the fifth entry
		tmp <- splits[5] 
		offset <- as.real(substr(tmp,2,3)) + ( as.real(substr(tmp,5,6)) / 60.0 )
		if(substr(tmp,1,1) != "+"){
			offset = offset * -1 
		}
		utc_offset <- c(utc_offset, offset)
	}
	
	# build data frame
	tz <- data.frame(ZONE_NAME=zone_name,UTC_OFFSET=utc_offset)

	# all done
	return(tz)
}

#######################################################################################
# write2ncdf
# write rastered data to netCDF
#######################################################################################

# create netCDF variable
write2ncdf <- function( rast, rastered, fname_out, add=FALSE, varname="TimeZone",
						OnlyAddPos=FALSE ){

	# create/open netCDF file. 
	ncid <- create_ncdf ( rast, varname, fname_out, overwrite=!add )

	# grid dimensions
	nx   <- dim(rast)[2]
	ny   <- dim(rast)[1]
	nxy  <- nx*ny
	ntot <- 0

	# Write data:
	# ... entire array at once ...
	if( TRUE ){
		outarr <- array(0,dim=c(nx,ny,1))
		outarr[,,1] <- outarr[,,1] + t(as.matrix((rastered)))[,ny:1]
		
		ncvar_put( ncid, varname, outarr )
		
	# ...or slice by slice.
	}else{
			
		# loop over all slices
		cnt  <- 0
		cat("Adding values to netCDF\n")
		cat("0 of",ny,"rows...\n")
		for(jj in 1:ny){
				
			# verbose
			cnt <- cnt + 1
			frc <- cnt/ny
			if( frc %% 0.01 == 0 ){
				cat( sprintf("%.2f",frc),"/ 1 completed...\n")
			}
			
			# get this slice
			ivals <- getValuesBlock(rastered, jj, 1, 1, nx)
			
			# TODO: replace with proper tz
			map_tz( ivals, tz )
			
			# number of non-zero values
			idx <- which(ivals > 0 )
			len <- length(idx)
			if( len == 0 ) next
			ntot <- ntot + len
			
			# add to netCDF
			
			# ncy is the lat coordinate in the netCDF file
			ncy  <- ny - jj + 1
			
			# add entire slice if specified so. This will overwrite any previously
			# set values.
			if( !OnlyAddPos ){
				ncvar_put( ncid, varname, ivals, start=c(1,ncy,1), count=c(nx,1,1) )
				
			# otherwise, add only non-zero (positive) values.
			}else{
			
				# verbose
				if(len>100){
					cat("   --> Adding", len, "values to column", jj, "\n")
				}
				
				# loop over all values
				for(ii in 1:len){
				
					# slice index
					iidx <- idx[ii]
					
					# current value
					val  <- ivals[iidx]
					
					# x-coordinate (lon) in netCDF file
					ncx  <- iidx
					
					# add to netCDF
					ncvar_put( ncid, varname, val, start=c(ncx,ncy,1), count=c(1,1,1) )
				}
			}
		}
		cat("number of values added:",ntot,"\n")	
	}

	# close netCDF
	nc_close(ncid)
	cat("data written to",fname_out,"\n")
}

#######################################################################################
# create_ncdf
# open/create netCDF file
#######################################################################################

# create netCDF variable
create_ncdf <- function( rast, varname, fname_out, overwrite=TRUE ){
	
	# if file already exists and overwrite is not set to true. Just read the
	# existing file. Data will be added to it.
	if( !overwrite && file.exists(fname_out) ){
		ncid <- nc_open( fname_out, write=TRUE )
		
		# read dummy value to make sure that variable exists.
		dum <- ncvar_get( ncid, varname, start=c(1,1,1), count=c(1,1,1) )
		
	# create file from scratch.
	}else{
	
		# get grid resolution
		xres <- res(rast)[1]
		yres <- res(rast)[2]
	
		# min and max boundaries are edges. construct mid-points from them
		x <- seq(xmin(rast)+xres/2,xmax(rast)-xres/2,by=xres)
		y <- seq(ymin(rast)+yres/2,ymax(rast)-yres/2,by=yres)
	
		# define coordinates
		lon  <- ncdim_def( "lon","degrees_east",  round(as.double(x),4) )
		lat  <- ncdim_def( "lat","degrees_north", round(as.double(y),4) )
		time <- ncdim_def( "time","days since 2000-01-01 00:00:00", 0, unlim=TRUE )
	
		# define variable
		ncvar <- ncvar_def( varname, "hours", list(lon,lat,time), 
							longname="UTC offset in hours", prec="float", 
							missval=-999.0, compression=1 )
	
		ncid  <- nc_create( fname_out, list(ncvar) )
	}
	
	# all done
	return( ncid )
}

#
##
### eof ###
